#Copyright Massachusetts Institute of technology, 1998.
#Written by Eric Prud'hommeaux for the World Wide Web Consortium

use strict;
require Exporter;
#require AutoLoader;

# use W3C::Database::DBIInterface;

$W3C::Rdf::Algernon::REVISION = '$Id: Algernon.pm,v 1.9 1999/12/14 13:14:39 eric Exp $ ';

package W3C::Rdf::Algernon;
use vars qw($VERSION $DSLI @ISA @EXPORT @EXPORT_OK @TODO);

@ISA = qw(W3C::Database::DBIRef Exporter); # AutoLoader);
@EXPORT = qw();
@EXPORT_OK = qw();
$VERSION = 0.11;
$DSLI = 'adpO';
@TODO = ('');

package W3C::Rdf::Algae::QueryPiece;
use vars qw($VARNAME $INDEX $DATA $DEPENDENTS);
($VARNAME, $INDEX, $DATA, $DEPENDENTS) = (0..3);

sub new {
    my ($proto, $partOfSpeach, $litteral, $varName, $varIndex, $newSymbol) = @_;
    my $class = ref($proto) || $proto;
    my $self = {};
    bless ($self, $class);
    ($self->{PARTOFSPEACH}, $self->{LITTERAL}, $self->{VARNAME}, $self->{VARINDEX}, $self->{NEWSYMBOL}) = 
	($partOfSpeach, $litteral, $varName, $varIndex, $newSymbol);
    return $self;
}

sub iterants (\@) {
    my ($self) = @_;
    return $self->{LITTERAL} if ($self->{LITTERAL});
    return undef if ($self->{NEWSYMBOL});
    # grab contextual values from $pResults at depth VARINDEX
    return @{$self->{CONTEXT}[-1][$DATA]};
}

sub absorb ($\@) {
    my ($self, $triple, $curContext) = @_;
    return 1 if (!$self->{NEWSYMBOL});
    my $absorbed = $self->{PARTOFSPEACH} == 0 ? $triple->getPredicate : 
	$self->{PARTOFSPEACH} == 1 ? $triple->getSubject : 
	    $triple->getObject;
    my $context = $curContext; # [@{$self->{CONTEXT}}];
    push (@{$context->[-1][$DATA]}, $absorbed);
    return 1;
}

sub deleteRow (\@) {
    my ($self, $pResults) = @_;
    for (my $i = @{$self->{CONTEXT}}; $i > 0; $i--) {
	print 'context: ('.join (',', @{$self->context}).')'."\n";
    }
    return 1;
}

sub show () {
    my ($self) = @_;
    my $ret;
    $ret .= $self->partOfSpeachStr.'  ';
    $ret .= 'litteral: '.$self->litteral->show.'  ' if ($self->litteral);
    $ret .= 'varName: '.$self->varName.'  ' if ($self->varName);
    $ret .= 'varIndex: '.$self->varIndex.'  ' if (defined $self->varIndex);
    $ret .= 'newSymbol: '.$self->newSymbol if ($self->newSymbol);
    $ret .= "\n";
    return $ret;
}

sub showContext () {
    my ($self, $start, $depth) = @_;
    my $leader = '    ' x $depth;
    my $ret = $leader.'['.$self->{VARNAMES}[$depth]."\n";
    foreach my $element (@$start) {
	if (ref $element eq 'ARRAY') {
	    $ret .= $self->showContext($element, $depth+1)."\n";
	} elsif (ref $element) {
	    $ret .= $leader.$element->show."\n";
	} elsif (defined $element) {
	    $ret .= $leader.'?'.$element."\n";
	} else {
	    $ret .= $leader.'undef'."\n";
	}
    }
    return $ret.$leader.']';
}

sub partOfSpeach () {return $_[0]->{PARTOFSPEACH};}
sub partOfSpeachStr () {my @posz = ('predicate', 'subject', 'object'); return $posz[$_[0]->{PARTOFSPEACH}];}
sub litteral () {return $_[0]->{LITTERAL};}
sub context () {return $_[0]->{CONTEXT};}
sub varName () {return $_[0]->{VARNAME};}
sub varIndex () {return $_[0]->{VARINDEX};}
sub newSymbol () {return $_[0]->{NEWSYMBOL};}

package W3C::Rdf::Algernon;

sub guessQuery {
    my ($self, $query, $baseUri) = @_;
    if ($query =~ m/^\s*\(/) {
	return $self->algernon($query, $baseUri);
    } elsif ($query =~ m/^\s*\{/) {
	my (@asks, @collects, %variables);
	while ($query =~ m/\G \s* \{ ([^\}]+) \} \s* \( \s* ([^\,]+) \s* \, \s* ([^,]+) \s* \) [\,\.]? /gcxs) {
	    my ($predicate, $subject, $object) = ($1, $2, $3);
	    push (@asks, '('.$predicate.' ?'.$subject.' ?'.$object.')');
	    if (!exists $variables{$subject}) {
		$variables{$subject} = @collects;
		push (@collects, '?'.$subject);
	    }
	    if (!exists $variables{$object}) {
		$variables{$object} = @collects;
		push (@collects, '?'.$object);
	    }
	}
	my $algaeQuery = "(ask '(".join ("\n       ", @asks).")\n :collect '(".join (' ', @collects).'))';
	my ($ret, $labels, $messages, $statements) = $self->algernon($algaeQuery, $baseUri);
	return ($ret, $labels, [$query, "- goes to -", $algaeQuery, @$messages], $statements);
    } else {
	die "4TF is \"$query\"?\n";
    }
}

sub algernon {
    my ($self, $algae, $baseUri) = @_;
    if (1) { # !defined $view) {
	# parse s-expressions into a structure
	my @structure = $self->parseAlgernon(\$algae);

	# be merciful - strip outer nested parens
	@structure = @{$structure[0]->content} while (@structure == 1 && ref $structure[0]->content eq 'ARRAY');

	return $self->interpretAlgernon(\@structure, $baseUri);
    }
}

sub W3C::SExpr::Atom::new (;\%) {
    my ($proto, $datum, $flags) = @_;
    my $class = ref($proto) || $proto;
    my $self = {$flags ? %$flags : ()}; # just copy them for now
    bless($self, $class);
    $self->{DATUM} = $datum;
    return $self;
}

sub W3C::SExpr::Atom::show (;\%) {
    my ($self, $flags) = @_;
    my $ret;
    $ret .= $self.' ' if ($flags->{-showRef});
    $ret .= 'quoted ' if ($self->{-quoted});
    $ret .= $self->{DATUM};
    return $ret;
}

sub W3C::SExpr::Atom::content () {return shift->{DATUM};}

@W3C::SExpr::Atom::String::ISA = qw(W3C::SExpr::Atom);
@W3C::SExpr::Atom::Symbol::ISA = qw(W3C::SExpr::Atom);
@W3C::SExpr::Atom::SExpr::ISA = qw(W3C::SExpr::Atom);
sub W3C::SExpr::Atom::SExpr::show (;\%) {
    my ($self, $flags) = @_;
    my $ret;
    $ret .= $self.' ' if ($flags->{-showRef});
    $ret .= ' quoted' if ($self->{-quoted});
    $ret .= ' ('.join (' ,', @{$self->{DATUM}->show}).')';
    return $ret;
}

sub parseAlgernon ($) {
    my ($self, $str) = @_;
    my @ret;
#    while ($$str !~ m/ \G \Z /gcxsi) {
    while (pos ($$str) < length ($$str)) {
	my ($object);
	if ($$str =~ m/ \G (\')? \( \s* /gcsxi) {
	    $object = new W3C::SExpr::Atom::SExpr($self->parseAlgernon($str));
	} elsif ($$str =~ m/ \G (\')? \"(.*?)\" \s* (\s | \)) \s* /gcsxi) {
	    $object = new W3C::SExpr::Atom::String($2);
	} elsif ($$str =~ m/ \G (\')? (.*?) \s* (\s | \)) \s* /gcsxi) {
	    $object = new W3C::SExpr::Atom::Symbol($2) if ($2);
	} else {
	    die 'don\'t know how to parse "'.substr($$str, pos $$str).'"';
	}
	$object->{-quoted} = 1 if ($1 eq '\'');
	push (@ret, $object) if ($object);
	return \@ret if ($3 eq ')');
    }
    return @ret;
}

use vars qw($VARNAME $INDEX $DATA $DEPENDENTS);
($VARNAME, $INDEX, $DATA, $DEPENDENTS) = (0..3);

sub interpretAlgernon (\@) { # for Algernon
    my ($self, $structure, $baseUri) = @_;
    my $directive = shift (@$structure);
    my @messages;
    my $view = $self->{TRIPLES};
    ($self->{RESULTS}, $self->{NEXT_RESULTS}, $self->{STATEMENTS}, $self->{NEXT_STATEMENTS}, 
     $self->{VAR_INDEX}, $self->{INDEX_VAR}, $self->{FIRST_PASS}) = ([], [], [], [], {}, [], 1);
    if ($directive->content eq 'ask') {
	my $pattern = shift (@$structure) ->content;
	for (my $i = 0; $i < @$pattern; $i++) {
	    my ($p, $s, $o) = @{$pattern->[$i]->content};
	    push (@messages,  'p:'.$p->show.' s:'.$s->show.' o:'.$o->show);

	    my @queryPieces;
	    $queryPieces[0] = $self->parseTripleList(0, $p, $baseUri);
	    $queryPieces[1] = $self->parseTripleList(1, $s, $baseUri);
	    $queryPieces[2] = $self->parseTripleList(2, $o, $baseUri);
	    my @iterOrder = (@queryPieces);

	    for (my $iRow = 0; $self->{FIRST_PASS} || $iRow < @{$self->{RESULTS}}; $iRow++) {
		my $row = $self->{FIRST_PASS} ? undef : $self->{RESULTS}[$iRow];
		my @lookfor = ($self->lookfor($queryPieces[0], $row), $self->lookfor($queryPieces[1], $row), $self->lookfor($queryPieces[2], $row));
		my @tmp = $self->triplesMatching($view, [\@lookfor]);
		my @tmp2;
		if (@tmp) {
		    foreach my $triple (@tmp) {

			# trim for cases like (?a ?a "blue")
			next if ($queryPieces[0]->varName && 
				 $queryPieces[0]->varName eq $queryPieces[1]->varName && 
				 $triple->getPredicate != $triple->getSubject);
			next if ($queryPieces[0]->varName && 
				 $queryPieces[0]->varName eq $queryPieces[2]->varName && 
				 $triple->getPredicate != $triple->getObject);
			next if ($queryPieces[1]->varName && 
				 $queryPieces[1]->varName eq $queryPieces[2]->varName && 
				 $triple->getSubject != $triple->getObject);

#			print $triple->show."\n";
			push (@tmp2, $triple);
		    }
		}
		if (@tmp2) {
		    for (my $iQueryResult = 0; $iQueryResult < @tmp2; $iQueryResult++) {
			# build a new array of results (and supporting statements) in a new array
			# This could be done with an array splice, but it's less efficient and less portable to C/Java.
			push (@{$self->{NEXT_RESULTS}}, $self->{FIRST_PASS} ? [] : [@$row]);
			push (@{$self->{NEXT_STATEMENTS}}, $self->{FIRST_PASS} ? [] : [@{$self->{STATEMENTS}[$iRow]}]);
			my $triple = $tmp2[$iQueryResult];
			push (@{$self->{NEXT_STATEMENTS}[-1]}, $triple);
			$self->absorb($iterOrder[0], $triple, $iQueryResult);
			$self->absorb($iterOrder[1], $triple, $iQueryResult);
			$self->absorb($iterOrder[2], $triple, $iQueryResult);
#print $iterOrder[2]->show."\n";
#print &showResults($results->[0], 0)."\n";
#&getMatrix($results->[0], \%varIndexes);
#print &showArray($results)."\n";
		    }
		} else {
		    # eliminate tied variables
    		}
		$self->{FIRST_PASS} = 0;
	    }
	    $self->{RESULTS} = $self->{NEXT_RESULTS};
	    $self->{NEXT_RESULTS} = [];
	    $self->{STATEMENTS} = $self->{NEXT_STATEMENTS};
	    $self->{NEXT_STATEMENTS} = [];
#	    print map {'('.join (' ', map {$_->show} @$_).")\n"} (@{$self->{RESULTS}}); print "\n";
	}
#	print "\n------\n"; print map {'('.join (' ', map {$_->show} @$_).")\n"} (@{$self->{RESULTS}}); print "\n";
	my $command = shift (@$structure) ->content;
	if ($command eq ':collect') {
	    my (@labels, @selects);
	    my $atom = shift (@$structure) ->content;
	    if (ref $atom eq 'ARRAY') {
#		push (@selects, '"("');
		foreach my $atom (@$atom) {
		    if ($atom->isa(qw(W3C::SExpr::Atom::Symbol))) {
			$atom = $atom->content;
			if ($atom =~ m/^\?(.*)/) {
			    $atom = $1;
			    if (defined $self->{VAR_INDEX}{$atom}) {
				push (@selects, $self->{VAR_INDEX}{$atom});
				push (@labels, $atom);
			    } else {
				die 'undefined symbol "'.$atom.'".';
			    }
			} elsif ($atom =~ m/^(\w+)\:\/\/([\w\d\-]+)/) {
			    $self->absoluteLink($baseUri, \$atom);
			    push (@selects, $atom);
			    push (@labels, $atom);
			} else {
			    die 'what kind of symbol is "'.$atom.'"?';
			}
		    } elsif ($atom->isa(qw(W3C::SExpr::Atom::String))) {
		    } else {
			die "what to do with \"$atom\"?";
		    }
		}
#		push (@selects, '")"');
	    } elsif ($atom =~ m/^\?(.*)/) {
		$atom = $1;
		if (defined $self->{VAR_INDEX}{$atom}) {
		    push (@selects, $self->{VAR_INDEX}{$atom});
		    push (@labels, $atom);
		} else {
		    die 'undefined symbol "'.$atom.'".';
		}
	    } elsif ($atom =~ m/^(\w+)\:\/\/([\w\d\-]+)/) {
		$self->absoluteLink($baseUri, \$atom);
		push (@selects, $atom);
		push (@labels, $atom);
	    } else {
		die 'what kind of symbol is "'.$atom.'"?';
	    }
	    my (@ret, @statements);
	    for (my $i = 0; $i < @{$self->{RESULTS}}; $i++) {
		my $result = $self->{RESULTS}[$i];
		my @row;
		foreach my $select (@selects) {
		    push (@row, $result->[$select]);
		}
		push (@ret, \@row);
		push (@statements, $self->{STATEMENTS}[$i]);
	    }
	    return (\@ret, \@labels, \@messages, \@statements);
	} else {
	    die "this interpreter only takes ':collect' as an 'ask' command.";
	}
#	return ([[1, 2, 3], ['e', 'z', 'as'], ['a', 'b', 'c']], \@messages);
    } else {
	die "this interpreter only supports 'ask'.";
    }
}

sub parseTripleList ($$\@\%$) {
    my ($self, $ordinal, $atom, $baseUri) = @_;
    if ($atom->isa(qw(W3C::SExpr::Atom::String))) {
	$atom = $atom->content;
	$self->absoluteString(\$atom);
	return new W3C::Rdf::Algae::QueryPiece($ordinal, $atom, undef, undef, 0);
    } elsif ($atom->isa(qw(W3C::SExpr::Atom::Symbol))) {
	$atom = $atom->content;
	if ($atom =~ m/^\?(.*)/) {
	    $atom = $1;
	    if (exists $self->{VAR_INDEX}{$atom}) {
#		return ([keys %{$results->[$varIndexes->{$atom}]}], $atom, -1);
		return new W3C::Rdf::Algae::QueryPiece($ordinal, undef, $atom, $self->{VAR_INDEX}{$atom}, 0);
	    } else {
		push (@{$self->{INDEX_VAR}}, $atom);
		$self->{VAR_INDEX}{$atom} = @{$self->{INDEX_VAR}}-1;
		return new W3C::Rdf::Algae::QueryPiece($ordinal, undef, $atom, $self->{VAR_INDEX}{$atom}, 1);
	    }
	} elsif ($atom =~ m/^(\w+)\:\/\/([\w\d\-]+)/) {
	    $self->absoluteLink($baseUri, \$atom);
	    return new W3C::Rdf::Algae::QueryPiece($ordinal, $atom, undef, undef, 0);
	} else {
	    die 'what kind of symbol is "'.$atom.'"?';
	}
    } else {
	die 'internal error: got a "'.$atom.'"';
    }
}

sub lookfor ($) {
    my ($self, $queryPiece, $row) = @_;
    if ($queryPiece->litteral) {
	return $queryPiece->litteral;
    } elsif ($queryPiece->newSymbol) {
	return undef;
    } else {
	return $row->[$queryPiece->varIndex];
    }
}

sub absorb ($$$) {
    my ($self, $queryPiece, $triple, $ordinal) = @_;
    if (defined $queryPiece->varIndex) {
	my $absorbed = $queryPiece->{PARTOFSPEACH} == 0 ? $triple->getPredicate : 
	    $queryPiece->{PARTOFSPEACH} == 1 ? $triple->getSubject : 
		$triple->getObject;
	$self->{NEXT_RESULTS}[-1][$queryPiece->varIndex] = $absorbed;
    }
}

sub showArray () {
    my ($start, $depth) = @_;
    my $leader = '    ' x $depth;
    my @ret;
    push (@ret, $leader.'[');
    foreach my $element (@$start) {
	if (ref $element eq 'ARRAY') {
	    push (@ret, &showArray($element, $depth+1));
	} elsif (ref $element) {
	    push (@ret, $leader.' '.$element->show);
	} elsif (defined $element) {
	    if ($element =~ m/^\d+$/) {
		push (@ret, $leader.' '.$element.'');
	    } else {
		push (@ret, $leader.' "'.$element.'"');
	    }
	} else {
	    push (@ret, $leader.' undef');
	}
    }
    push (@ret, $leader.']');
    return join ("\n", @ret);;
}

sub getMatrix () {
    my ($start, $varIndexes) = @_;
    my @matrix = ([]);
    &gatherMatrix($start, 0, \@matrix);
    return @matrix;
}

sub gatherMatrix () {
    my ($start, $depth, $matrix) = @_;
    for (my $i = 0; $i < @{$start->[$DATA]}; $i++) {
	my $element = $start->[$DATA][$i];
	my $elementStr;
	if (ref $element) {
	    $elementStr = '"'.$element->show.'"';
	} elsif ($element =~ m/^\d+$/) {
	    $elementStr = $element;
	} else {
	    $elementStr = '"'.$element.'"';
	}
	if ($i > 0) {
	    push (@$matrix, $depth ? $matrix->[-1] : []);
	}
	$matrix->[-1][$depth] = $elementStr;
	my $dependents = $start->[$DEPENDENTS][$i];
	if ($dependents && @$dependents) {
	    &gatherMatrix($dependents, $depth+1, $matrix);
	}
    }
}

sub showResults () {
    my ($start, $depth) = @_;
    my $leader = '    ' x $depth;
    my @ret;
    push (@ret, $leader.'(');
    for (my $i = 0; $i < @{$start->[$DATA]}; $i++) {
	my $element = $start->[$DATA][$i];
	if (ref $element) {
	    push (@ret, $leader.' "'.$element->show.'"');
	} elsif ($element =~ m/^\d+$/) {
	    push (@ret, $leader.' '.$element);
	} else {
	    push (@ret, $leader.' "'.$element.'"');
	}
	my $dependents = $start->[$DEPENDENTS][$i];
	if ($dependents && @$dependents) {
	    push (@ret, &showResults($dependents, $depth+1));
	}
    }
    push (@ret, $leader.')');
    return join ("\n", @ret);;
}

1;


__END__

=head1 NAME

W3C::Rdf::RdfParser - parse RDF/XML and fill an RdfDB

=head1 SYNOPSIS

    use strict;

    BEGIN {unshift@INC,('../../..');}

    use W3C::Rdf::MysqlDB;

    my ($query) = @_;
    my $dbConf;
    use W3C::Util::Properties;
    $dbConf = new W3C::Util::Properties('../../../Conf/MysqlDB.prop');
    my $rdfDB = new W3C::Rdf::MysqlDB({-errorHandler => undef, -lazyReification => 1, -properties => $dbConf});

    $query = "(ask '((http://www.w3.org/schema/certHTMLv1/access ?x \"acc1\") 
(?y ?x \"http://user1\")) 
:collect ?x)" if (!$query);
    my ($nodes2, $selects, $messages) = $rdfDB->algernon($query, $revision);

    print join ("\n", @$messages);
    print 'algernon "'.$query.'"';
    print @$nodes2 > 1 ? "\n--------------------------------\n" : ' -> ';
    print join ("\n", map {join (', ', map {$_->show} @$_)} @$nodes2)."\n";

=head1 DESCRIPTION

perform algernon-like queries on an RDF database

This module is part of the W3C::Rdf CPAN module.

=head1 AUTHOR

Eric Prud\'hommeaux <eric@w3.org>

=head1 SEE ALSO

W3C::Rdf::RdfDB(3) W3C::Rdf::RdfParser(3) perl(1).

=cut
