#Copyright Massachusetts Institute of technology, 1998.
#Written by Eric Prud'hommeaux for the World Wide Web Consortium

use strict;
require Exporter;
require AutoLoader;

# BEGIN {unshift@INC,('../..');}

$W3C::Rdf::RdfApp::REVISION = '$Id: RdfApp.pm,v 1.17 1999/12/17 22:43:00 eric Exp $ ';

package W3C::Rdf::RdfApp;
use vars qw($VERSION $DSLI @ISA @EXPORT @EXPORT_OK);
@ISA = qw(Exporter AutoLoader);

use W3C::Rdf::RdfParser;
use W3C::SAX::XmlParser;
use W3C::SAX::InputSource;
use W3C::SAX::HandlerBase;
use W3C::SAX::ErrorHandlerImpl;
use W3C::Rdf::RdfDB;
@EXPORT = qw();
@EXPORT_OK = qw();
$VERSION = 0.94;
$DSLI = 'adpO';

#####
# Externally available constants
*W3C::Rdf::RdfApp::ONE = \ 'one';

sub new {
    my ($proto, $ARGV) = @_;
    my $class = ref($proto) || $proto;
    my $self = {};
    bless ($self, $class);
    return $self;
}

sub execute (\@) {
    my ($self, $argv) = @_;
    my @errors;
    $self->preArgumentDefaults($argv, \@errors);
    $self->parseArgumentArray($argv, \@errors);
    $self->postArgumentDefaults($argv, \@errors);
    $self->parse;
    $self->render;
}

sub preArgumentDefaults {
    my ($self, $errs) = @_;
    $self->{ARGS}{-props} = {};
    return 0;
}

sub parseArgumentArray {
    my ($self, $argv, $errs) = @_;
    my $errorCount = 0;
  NEXTARG:
    for (my $i = 0; $i < @$argv; $i++) {
	my $arg = $argv->[$i];
	if ($arg =~ m/\A @(.*) \Z/x) {
	    my $scriptFile;
	    if ($1) {
		$scriptFile = $1;
	    } elsif ($i < @$argv - 1) {
		$scriptFile = $argv->[++$i];
	    } else {
		$errorCount++;
	    }
	    open (SCRIPT_FILE, '<'.$scriptFile) || die "error opening \"$scriptFile\": $!";
	    local $/ = undef;
	    my $script = <SCRIPT_FILE>;
	    close(SCRIPT_FILE);
	    my ($line, @parms);
	    while (pos $script < length $script && $script =~ m/\G ([^\s\\\'\"]*) ([\s\\\'\"])? /gcxs) {
		my ($word, $delim) = ($1, $2);
		$line .= $word;
		if ($delim =~ m/\s/) {
		    push (@parms, $line) if ($line ne '$*');
		    $line = undef;
		    if ($script =~ m/\G \s* /gcxs) {}
		} elsif ($delim eq '\\') {
		    if ($script =~ m/\G (.) /gcxs) {$line .= $1 if ($1 ne "\r" && $1 ne "\n");}
		} elsif ($delim eq '\'') {
		    while ($script =~ m/\G ([^\\\']*) ([\\\'\z]) /gcxs) {
			$line .= $1;
			if ($2 eq '\\') {
			    if ($script =~ m/\G (.) /gcxs) {$line .= $1 if ($1 ne "\r" && $1 ne "\n");}
			} elsif ($2 eq '\'') {
			    last;
			}
		    }
		} elsif ($delim eq '"') {
		    while ($script =~ m/\G ([^\\\"]*) ([\\\"\z]) /gcxs) {
			$line .= $1;
			if ($2 eq '\\') {
			    if ($script =~ m/\G (.) /gcxs) {$line .= $1 if ($1 ne "\r" && $1 ne "\n");}
			} elsif ($2 eq '"') {
			    last;
			}
		    }
		} else {
		    push (@parms, $line) if (defined $line && $line ne '$*');
		    $line = undef;
		    # \z so at end of input - hmm - can't count on that. bummer
		}
	    }
	    push (@parms, $line) if ($line);
	    my $zero = shift (@parms);
	    splice (@$argv, $i, 1, @parms);
	    $i--;
	} elsif ($arg =~ m/\A -n(.*) \Z/x) {
	    if ($1) {
		$self->{ARGS}{'-rdfUri'} = $1;
	    } elsif ($i < @$argv - 1) {
		$self->{ARGS}{'-rdfUri'} = $argv->[++$i];
	    } else {
		$errorCount++;
	    }
	} elsif ($arg =~ m/\A -a(.*) \Z/x) {
	    if ($1) {
		push (@{$self->{ARGS}{'-algae'}}, $1);
	    } elsif ($i < @$argv - 1) {
		push (@{$self->{ARGS}{'-algae'}}, $argv->[++$i]);
	    } else {
		$errorCount++;
	    }
	} elsif ($arg =~ m/\A \>(.*) \Z/x) {
	    if ($1) {
		open (OUT, '>'.$1) || die $!;
		$self->{OUT_FILE} = *OUT;
	    } elsif ($i < @$argv - 1) {
		open (OUT, '>'.$argv->[++$i]) || die $!;
		$self->{OUT_FILE} = *OUT;
	    } else {
		$errorCount++;
		push(@$errs, 'unknown arg: "'.$arg.'"');
	    }
	} elsif ($arg =~ m/\A -d(.*) \Z/x) {
	    if ($1) {
		$self->{ARGS}{'-rdfDB'} = $1; #  eq 'object';
	    } elsif ($i < @$argv - 1) {
		$self->{ARGS}{'-rdfDB'} = $argv->[++$i]; #  eq 'object';
	    } else {
		$errorCount++;
	    }
	} elsif ($arg =~ m/\A -set(.*) \Z/x) {
	    my $parm;
	    if ($1) {
		$parm = $1;
	    } elsif ($i < @$argv - 1) {
		$parm = $argv->[++$i];
	    } else {
		$errorCount++;
	    }
	    if ((my ($attr, $value) = split ('\s*=\s*', $parm)) == 2) {
		$self->{ARGS}{-props}{$attr} = $value;
	    } else {
		$errorCount++;
	    }
	} elsif ($arg =~ m/\A ( -s ) \Z/x) {
	    $self->{ARGS}{'-sourceOnly'} = 1;
	} elsif ($arg =~ m/\A ( -ByPredicate ) \Z/x) {
	    $self->{ARGS}{'-startState'} = $W3C::Rdf::RdfParser::MODE_ByPredicate;
	} elsif ($arg =~ m/\A ( -BySubjectType ) \Z/x) {
	    $self->{ARGS}{'-startState'} = $W3C::Rdf::RdfParser::MODE_BySubjectType;
	} elsif ($arg =~ m/\A ( -BySubject ) \Z/x) {
	    $self->{ARGS}{'-startState'} = $W3C::Rdf::RdfParser::MODE_BySubject;
	} elsif ($arg =~ m/\A ( -{0,2}help | -{0,1}\? ) \Z/x) { # help -help --help ? -?
	    $self->{ARGS}{'-help'} = 1;
	} elsif ($arg =~ m/\A ( -{1,2}verbose | -{1,2}v ) \Z/x) { # -verbose --verbose -v --v
	    $self->{ARGS}{'-verbose'} = 1;
	} else {
	    push (@{$self->{ARGS}{-files}}, $arg);
	}
    }

    return $errorCount;
}

sub postArgumentDefaults {
    my ($self, $errs) = @_;
    return 0;
}

sub parse () {
    my ($self) = @_;
    $self->{XML_PARSER} = new W3C::SAX::PerlXmlParser;
#    $self->{XML_PARSER} = new W3C::SAX::ExpatXmlParser;
    my $errorHandler = new W3C::SAX::ErrorHandlerImpl({-documentLocator => $self->{XML_PARSER}});

    # pick an RDF database
    if ($self->{ARGS}{-rdfDB} eq 'W3C::Rdf::MysqlDB') {
	require W3C::Rdf::MysqlDB;
	require W3C::Util::Properties;
	my $dbConf = new W3C::Util::Properties('../../../Conf/MysqlDB.prop');
	$self->{RDF_DB} = new W3C::Rdf::MysqlDB({-errorHandler => $errorHandler, -lazyReification => 1, -properties => $dbConf});
    } else {
	use W3C::Rdf::RdfDB;
	$self->{RDF_DB} = new W3C::Rdf::RdfDB({-errorHandler => $errorHandler, -lazyReification => 1});
    }

    $self->{RDF_DB}->setRdfPrefix($self->{ARGS}{-rdfUri}) if ($self->{ARGS}{-rdfUri});
    $self->{RDF_PARSER} = new W3C::Rdf::RdfParser({-nextTrustControlBlock=>1, -trustedIds=>{}, -untrustedIds=>{}, 
						   -rdfDB=>$self->{RDF_DB}, -namespaceHandler=>undef, 
						   -errorHandler=>$errorHandler, -appendAttribution=>0});

    # it's easiest to stick all your document handlers in a W3C::SAX::HandlerList
    my $handlerList = new W3C::SAX::HandlerList({-documentHandlers => [$self->{RDF_PARSER}], 
						 -errorHandlers => [$errorHandler]});

    # use a W3C::SAX::NamespaceHandler stream to map from prefix:tag - pass to rdfParser
    # so it translate back for pretty-printing
    $self->{NAMESPACE_HANDLER} = new W3C::SAX::NamespaceAndStringHandler({-documentHandler => $handlerList, 
									  -errorHandler => $handlerList});
    $self->{RDF_PARSER}->setNamespaceHandler($self->{NAMESPACE_HANDLER});

    # and have the W3C::SAX::HandlerList be the document handler and the error handler
    $self->{XML_PARSER}->setDocumentHandler($self->{NAMESPACE_HANDLER});
    $self->{XML_PARSER}->setErrorHandler($handlerList);

    # assign a URI to the XML in InputSource and parse it
    foreach my $source (@{$self->{ARGS}{-files}}) {
	use W3C::Util::Exception;
	my $inputSource;
	eval {
	    $inputSource = new W3C::SAX::FileInputSource($source);
	    $self->{NAMESPACE_HANDLER}->addResourceMap($inputSource->getPublicId);
	    $self->{RDF_PARSER}->setState($self->{ARGS}{'-startState'}) if (defined $self->{ARGS}{'-startState'});
	    $self->{XML_PARSER}->parse($inputSource);
	    $self->{ATTRIBUTIONS}{$inputSource->getPublicId} = $self->{RDF_PARSER}->getRootAttribution;
#	    print join (' | ', ($self->{NAMESPACE_HANDLER}->getUnknownNamespaces))."\n";
	    if (defined $self->{NAMESPACE_HANDLER}->getUnknownNamespaceDeclaration('rdf')) {
		warn "rdf: namespace undeclared: assuming ".$self->{RDF_DB}->getRdfPrefix."
-- fix this by adding 'xmlns:rdf=\"".$self->{RDF_DB}->getRdfPrefix."\"' to rdf:".$self->{NAMESPACE_HANDLER}->getUnknownNamespaceDeclaration('rdf')."\n";
		$self->{NAMESPACE_HANDLER}->addNamespace('rdf', $self->{RDF_DB}->getRdfPrefix, $inputSource->getPublicId);
	    }
	    $self->{RDF_PARSER}->append;
	}; if ($@) {if (my $ex = &catch('W3C::Util::FileNotFoundException')) {
	    warn "Can't open input file \"".$ex->getFilename."\" - skipping\n";
	} elsif (my $ex = &catch('W3C::SAX::SAXParseException')) {
	    warn $ex->getMessage.' - abandoning "'.$inputSource->getPublicId."\".\n";
	} elsif (my $ex = &catch('W3C::Util::Exception')) {
	    $ex->printStackTrace;
	    exit;
	} else {die $@;}}
    }
}

sub render () {
}

1;

__END__

=head1 NAME

W3C::Rdf::RdfApp - Common handlers and CLI for RDF apps

=head1 SYNOPSIS

    use W3C::Rdf::RdfApp;
    package testRdfParser;
    @testRdfParser::ISA = qw(W3C::Rdf::RdfApp);

    my $tester = new testRdfParser;
    $tester->execute(\@ARGV);

    sub render {
	my ($self) = @_;
	$query = "(ask '((http://www.w3.org/schema/certHTMLv1/access ?x \"acc1\") 
(?y ?x \"http://user1\"))
:collect ?x)" if (!$query);
	my ($nodes2, $messages) = $self->{RDF_DB}->algernon($query, $self->{RDF_PARSER}->{REVISION});
    }

=head1 DESCRIPTION

Common handlers and CLI for RDF apps.

This module is part of the W3C::RDF CPAN module suite.

=head1 AUTHOR

Eric Prud\'hommeaux <eric@w3.org>

=head1 SEE ALSO

W3C::Rdf::RdfParser(3) W3C::Rdf::RdfDB(3) W3C::SAX::XmlParser(3) perl(1).

=cut
