#Copyright Massachusetts Institute of technology, 1998.
#Written by Eric Prud'hommeaux for the World Wide Web Consortium

# structure:
# RdfVisualizer holds n WindowSets
# WindowSet has:
#   ViewMaster    - selection of current view
#   QueryMaster   - selection of current query
#   StatementList ClientWindow for displaying current view of statements
#   Graph         ClientWindow for rendering highighted statement
#   QueryResults  ClientWindow table view of highlighted query 

# action maps:

# user enters a source view:
# ViewMaster::onReturn
#  if (!view = cache{sourceString})
#    view = new SourceView
#    ViewMaster::add(view)
#  Statements::refreshList
#   SourceView::getStatements

# user re-selects a source view:
# QueryMaster::onArrow
#  QueryView::activate(QueryMaster)
#    QueryView->imActive(QueryView)
#  Statements::refreshList...

# user enters a query:
# QueryMaster::onReturn
#  if (!view = cache{queryString})
#    view = new QueryView
#    QueryMaster::add(view)
#    ViewMaster::add(view)
#  Statements::refreshList
#   QueryView::getStatements
#    Results::addResults
#     resultsView = new ResultsView

# user re-selects a query view:
# QueryMaster::onArrow
#  QueryView::activate(QueryMaster)
#    QueryView->imActive(QueryView)
#  Statements::refreshList...

# user re-selects a query:
# ViewMaster::onArrow
#  QueryView::activate(ViewMaster)
#    QueryView->imActive(QueryView)
#  Statements::refreshList...

# user selects a query result
# QueryResults::onReturn
#  ResultsView::activate(QueryResults)
#   ViewMaster->imActive(ResultsView)
#  Statements::refreshList

use strict;
#require Exporter;
#require AutoLoader;
require Tk;

$W3C::Rdf::RdfVisualizer::REVISION = '$Id: RdfVisualizer.pm,v 1.29 1999/12/16 20:41:25 eric Exp $ ';

# inheritance (@ISAs) needs to precede the call to new
@W3C::Rdf::RdfVisualizer::StatementList::ISA = qw(W3C::Rdf::RdfVisualizer::ClientWindow);
@W3C::Rdf::RdfVisualizer::MainWindow::ISA = qw(W3C::Rdf::RdfVisualizer::ClientWindow);
@W3C::Rdf::RdfVisualizer::ListMaster::ISA = qw(W3C::Rdf::RdfVisualizer::ClientWindow);
@W3C::Rdf::RdfVisualizer::ViewMaster::ISA = qw(W3C::Rdf::RdfVisualizer::ListMaster);
@W3C::Rdf::RdfVisualizer::QueryMaster::ISA = qw(W3C::Rdf::RdfVisualizer::ListMaster);
@W3C::Rdf::RdfVisualizer::GraphMaster::ISA = qw(W3C::Rdf::RdfVisualizer::ListMaster);
@W3C::Rdf::RdfVisualizer::Graph::ISA = qw(W3C::Rdf::RdfVisualizer::MainWindow);
@W3C::Rdf::RdfVisualizer::QueryResults::ISA = qw(W3C::Rdf::RdfVisualizer::MainWindow);

@W3C::Rdf::RdfVisualizer::SourceView::ISA = qw(W3C::Rdf::RdfVisualizer::View);
@W3C::Rdf::RdfVisualizer::QueryView::ISA = qw(W3C::Rdf::RdfVisualizer::View);
@W3C::Rdf::RdfVisualizer::ProofView::ISA = qw(W3C::Rdf::RdfVisualizer::View);
@W3C::Rdf::RdfVisualizer::GraphView::ISA = qw(W3C::Rdf::RdfVisualizer::View);
@W3C::Rdf::RdfVisualizer::GraphStatementView::ISA = qw(W3C::Rdf::RdfVisualizer::GraphView);
@W3C::Rdf::RdfVisualizer::GraphNodeView::ISA = qw(W3C::Rdf::RdfVisualizer::GraphView);

package W3C::Rdf::RdfVisualizer;
use vars qw($VERSION $DSLI @ISA @EXPORT @EXPORT_OK @TODO);
@ISA = qw();
@EXPORT = qw();
@EXPORT_OK = qw();
$VERSION = 0.94;
$DSLI = 'adpO';
@TODO = ('hide internals of -namespaceHandler');

sub new {
    my ($proto, $parms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $parms;
    bless ($self, $class);
    push (@{$self->{-windowSets}}, new W3C::Rdf::RdfVisualizer::WindowSet({%$parms, -main => $self}));
    return $self;
}

sub cloneWindowSet {
    my ($self, $toClone) = @_;
    my $dolly = W3C::Rdf::RdfVisualizer::WindowSet::new('W3C::Rdf::RdfVisualizer::WindowSet', {%$toClone});
    $dolly->makeWindow($self->{MAIN_WINDOW});
    push (@{$self->{-windowSets}}, $dolly);
}

sub browse {
    my ($self) = @_;
    $self->{MAIN_WINDOW} = Tk::MainWindow->new(-title => $self->{-mainWinTitle});
    $self->{MAIN_WINDOW}->withdraw;
    foreach my $windowSet (@{$self->{-windowSets}}) {
	$windowSet->makeWindow($self->{MAIN_WINDOW});
    }
    &Tk::MainLoop;
}

sub notifyExit {
    my ($self, $doomed) = @_;
    $self->{-windowSets} = [grep {$_ != $doomed} @{$self->{-windowSets}}];
    $self->{MAIN_WINDOW}->destroy if (!@{$self->{-windowSets}});
}

package W3C::Rdf::RdfVisualizer::MatrixRenderer;
sub new {
    my ($proto, $parms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $parms;
    bless ($self, $class);
    $self->makeWidget;
    return $self;
}
# virtual functions
sub makeWidget {die}
sub pack {return shift->{-widget}->pack(@_);}
sub bind {return shift->{-widget}->bind(@_);}
sub size {return shift->{-widget}->size(@_);}
sub bell {return shift->{-widget}->bell(@_);}
sub highlight {
    my ($self, $index) = @_;
    $self->{-widget}->selection('clear', 0, 'end');
    return if (!defined $index); # filter out Graph::graphPick(undef) calls
    $self->{-widget}->selection('set', $index);
    $self->{-widget}->activate($index);
    $self->{-widget}->see($index);
}
sub startRows {
    my ($self) = @_;
    $self->{-data} = [];
    $self->{-widths} = [];
}
sub addRow {
    my ($self, $data) = @_;
    for (my $i = 0; $i < @$data; $i++) {
	my $width = length $data->[$i];
	$self->{-widths}[$i] = $width if ($self->{-widths}[$i] < $width);
    }
    push (@{$self->{-data}}, $data);
}
sub endRows {die}
sub destroy {
    my ($self) = @_;
    $self->{-widget}->destroy; # @@@ 4TF: why dan't I destroy this widget?
    # PORTNOTE: delete $self
}

package W3C::Rdf::RdfVisualizer::ListboxMatrixRenderer;
@W3C::Rdf::RdfVisualizer::ListboxMatrixRenderer::ISA = qw(W3C::Rdf::RdfVisualizer::MatrixRenderer);
sub makeWidget {
    my ($self) = @_;
    my $f = $self->{-parentWindow}->Font(-family => 'Courier');
    $self->{-widget} = $self->{-parentWindow}->Scrolled('Listbox', -scrollbars => 'osoe', -selectmode => 'extended', -font => $f);
    $self->{-widget}->bind("<Double-ButtonPress-1>", sub {$self->selectAction});
    $self->{-widget}->bind("<Return>", sub {$self->selectAction});
#   $self->{-widget}->bind("<Space>", [\&triplesPick, $self]);
}
sub selectAction {
    my ($self) = @_;
    &{$self->{-command}} ((($self->{-widget}->curselection)[0], 0));
}
sub startRows {
    my ($self) = @_;
    $self->SUPER::startRows;
    $self->{-widget}->delete(0, 'end');
}
sub endRows {
    my ($self) = @_;
    foreach my $row (@{$self->{-data}}) {
	my $text;
	for (my $j = 0; $j < @$row; $j++) {
	    my $column = $row->[$j];
	    $text .= $column;
	    $text .= ' ' x ($self->{-widths}[$j] - length $column);
	    $text .= ' ' if ($j < @$row-1);
	}
	$self->{-widget}->insert('end', $text);
    }
    my $windowWidth = 0;
    for (my $i = 0; $i < @{$self->{-widths}}; $i++) {
	my $width = $self->{-widths}[$i];
	$windowWidth += $width;
    }
    return $windowWidth;
}

package W3C::Rdf::RdfVisualizer::ColumnMatrixRenderer;
@W3C::Rdf::RdfVisualizer::ColumnMatrixRenderer::ISA = qw(W3C::Rdf::RdfVisualizer::MatrixRenderer);
sub get {
    my ($self) = (shift);
    my @array = $self->{-widget}->get(@_);
    my $tmp = $array[0];
    my $entry = join (' ', @$tmp);
    return $entry;
}
sub makeWidget {
    my ($self, $parms) = @_;
    $self->{-widget} = $self->{-parentWindow}->Columns('-columnlabels' => $self->{-columnLabels},
						       '-command' => $self->{-command},
						       '-listforeground' => 'blue4',
						       '-listbackground' => 'beige',
						       '-buttonbackground' => 'brown',
						       '-buttonforeground' => 'white',
						       '-selectmode' => 'extended',
						       '-sort' => 'false');
}
sub startRows {
    my ($self) = @_;
    $self->SUPER::startRows;
    $self->{-widget}->delete(0, 'end');
}
sub endRows {
    my ($self) = @_;
    foreach my $row (@{$self->{-data}}) {
	$self->{-widget}->insert('end', [@$row]);
    }
    my $windowWidth = 0;
    for (my $i = 0; $i < @{$self->{-widths}}; $i++) {
	my $width = $self->{-widths}[$i];
	$self->{-widget}->indexedbutton($i)->configure('-width' => $width-1);
	$windowWidth += $width;
    }
    return $windowWidth;
}

sub destroy {
    my ($self) = @_;
#    $self->{-widget}->destroy; # @@@ 4TF: why dan't I destroy this widget?
    # PORTNOTE: delete $self
}

package W3C::Rdf::RdfVisualizer::WindowSet;
sub new {
    my ($proto, $parms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $parms;
    bless ($self, $class);
    $self->{SHOW_REIFICATIONS} = 0;
    my $mwf;
    eval {
	die if ($self->{'noColumns'});
	require Tk::Columns;
	$mwf = sub {return new W3C::Rdf::RdfVisualizer::ColumnMatrixRenderer(@_);};
    }; if ($@) {
	$mwf = sub {return new W3C::Rdf::RdfVisualizer::ListboxMatrixRenderer(@_);};
    }
    $self->{STATEMENTS} = new W3C::Rdf::RdfVisualizer::StatementList({-windowSet => $self, 
								      -matrixWidgetFactory => $mwf});
    $self->{GRAPH} = new W3C::Rdf::RdfVisualizer::Graph({-windowSet => $self});
    $self->{QRESULTS} = new W3C::Rdf::RdfVisualizer::QueryResults({-windowSet => $self, 
								   -matrixWidgetFactory => $mwf});
    $self->{VIEW_MASTER} = new W3C::Rdf::RdfVisualizer::ViewMaster({-windowSet => $self});
    $self->{QUERY_MASTER} = new W3C::Rdf::RdfVisualizer::QueryMaster({-windowSet => $self});
    return $self;
}

sub getQueryOptions {
    my ($self) = @_;
    return defined $self->{-queryOptions} ? %{$self->{-queryOptions}} : ();
}

sub quit {exit;}

my ($X, $Y, $TS, $AW, $AH) = (300, 300, 10, 5, 8);

sub makeWindow {
    my ($self, $mw) = @_;
    my $mw = $mw->Toplevel;
    $mw->title('RDF Browser');
    $mw->protocol('WM_DELETE_WINDOW', sub {$self->destroyWindow});
    $self->{CONTROL_WINDOW} = $mw;
    $mw->bind("<Control-Q>", [\&quit]); # @@@ hmm - no worky, maybe I should pepper the canvas with quit buttons?

    $self->{GRAPH}->makeWindow($mw);
    $self->{QRESULTS}->makeWindow($mw);

    # top row is some input fields
    {
	my $mw = $mw->Frame(-relief => 'groove', -bd => 2)->pack(-fill => 'x', -side => 'top');
	$self->{VIEW_MASTER}->makeWindow($mw);
	$self->{QUERY_MASTER}->makeWindow($mw);
    }

    # bottom row is just the exit button
    {
	my $mw = $mw->Frame(-relief => 'groove', -bd => 2)->pack(-fill => 'x', -side => 'bottom');
	$mw->Button(-text => "clone", -underline => 1, -command => sub {$self->{-main}->cloneWindowSet($self);})
	    ->pack(-side => 'left');
	$mw->Button(-text => "exit", -underline => 1, -command => sub {$self->destroyWindow})->pack(-side => 'right');
    }

    # now the namespace options for each source
    foreach my $sourceId (keys %{$self->{-namespaceHandler}->{NAMESPACE_REVERSE}}) {
	$mw->Label(-text => 'show namespaces from source: '.$sourceId, -justify => 'left')->pack(-anchor => 'w');
	my $mw = $mw->Frame(-relief => 'groove', -bd => 2)->pack(-padx => 20);
	foreach my $expanded (keys %{$self->{-namespaceHandler}->{NAMESPACE_REVERSE}{$sourceId}}) {
	    my $ns = $self->{-namespaceHandler}->{NAMESPACE_REVERSE}{$sourceId}{$expanded};
	    $self->{ENS}{$sourceId}{$expanded} = 1;
	    $mw->Checkbutton(-text => "xmlns:$ns=\"$expanded\"", 
			     -anchor => 'e', 
			     -padx => 5, 
			     -variable => \ $self->{ENS}{$sourceId}{$expanded}, 
			     -command => sub {$self->nsPick})->pack(-anchor => 'w');
	}
    }

    # a new frame provides the body for the search prompt, triple list, and graph canvas
    {
	my $listFrame = $mw->Frame(-relief => 'groove', -bd => 2)->pack(-fill => 'both', -expand => 1);
	my $searchFrame;
	{

	    # framed search prompt on the top
	    $searchFrame = $listFrame->Frame(-relief => 'groove', -bd => 2)->pack(-fill => 'x');

	    # View options:
	    # ... node
	    $searchFrame->Checkbutton(-text => 'query results', 
			     -underline => 5, 
			     -variable => $self->{QRESULTS}->getShowVariable, 
			     -command => sub {$self->{QRESULTS}->viewWindow})->pack(-anchor => 'w', -side => 'right', -padx => 10);
	    # ... graph
	    $searchFrame->Checkbutton(-text => 'graph', 
			     -underline => 5, 
			     -variable => $self->{GRAPH}->getShowVariable, 
			     -command => sub {$self->{GRAPH}->viewWindow})->pack(-anchor => 'w', -side => 'right', -padx => 10);
	    # ... reifications
	    $searchFrame->Checkbutton(-text => 'reifications', 
			     -underline => 5, 
			     -variable => \ $self->{SHOW_REIFICATIONS}, 
			     -command => sub {$self->changeLazyReification})->pack(-anchor => 'w', -side => 'right', -padx => 10);
	    $searchFrame->Label(-text => 'View: ')->pack(-side => 'right');
	}

	$self->{STATEMENTS}->makeWindow($listFrame, $searchFrame);
    }

    $self->{VIEW_MASTER}->passed(join (' ', sort keys %{$self->{-attributionsByInputSource}}));
    $self->refreshList;
    foreach my $alga (@{$self->{-algae}}) {
	$self->{QUERY_MASTER}->passed($alga);
    }
}

sub destroyWindow {
    my ($self) = @_;
    $self->{CONTROL_WINDOW}->destroy();
    $self->{-main}->notifyExit($self)
}

sub changeLazyReification {
    my ($self) = @_;
    $self->{-RdfDB}->setLazyReification(!$self->{SHOW_REIFICATIONS});
    $self->refreshList;
    $self->graphPick($self->{GRAPH}->getLastCurrent);
}

sub nsPick {
    my ($self) = @_;
    $self->refreshList;
}

sub renderStatementNUKE {
    my ($self, $triple) = @_;
    my $text = $triple->show('', $self->{-namespaceHandler});
    my ($sourceId, $key, $predicateText) = 
	$self->{-namespaceHandler}->unmapNamespace($triple->getPredicate->show, $triple->getAttribution->getUri);
    return (!defined $sourceId && !defined $key || $self->{ENS}{$sourceId}{$key}, $text);
}

sub renderStatementSegments {
    my ($self, $statement) = @_;
    my ($sourceId, $key, $predicateText) = 
	$self->{-namespaceHandler}->unmapNamespace($statement->getPredicate->show, $statement->getAttribution->getUri);
    my @segments;
    my $enabled = !defined $sourceId && !defined $key || $self->{ENS}{$sourceId}{$key};
    if ($enabled) {
	push (@segments, $statement->getAttribution->show('', $self->{-namespaceHandler}));
	push (@segments, '(');
	push (@segments, $self->renderNode($statement->getPredicate, $statement->getAttribution));
	push (@segments, $self->renderNode($statement->getSubject, $statement->getAttribution));
	push (@segments, $self->renderNode($statement->getObject, $statement->getAttribution, {-stringTrunc => 30,
											       -truncFlag => '...'}));
	push (@segments, ')');
#	push (@segments, $statement->getReifiedId ? $self->renderNode($statement->getReifiedId, $statement->getAttribution) : '');
#	push (@segments, $statement->getReifiedIn ? $self->renderNode($statement->getReifiedIn, $statement->getAttribution) : '');
    }
    return ($enabled, \@segments);
}

sub renderNode {
    my ($self, $node, $attribution, $flags, $enabled) = @_;
    my $namespaceId = $attribution ? $attribution->getUri : undef;
    my ($sourceId, $key, $text) = 
	$self->{-namespaceHandler}->unmapNamespace($node->show(undef, undef, $flags), $namespaceId);
    if ($enabled) {
	$$enabled = !defined $sourceId && !defined $key || $self->{ENS}{$sourceId}{$key};
    }
    return $text;
}

# control window event redirections
sub addView {return shift->{VIEW_MASTER}->add(@_);}
sub getStatements {return shift->{VIEW_MASTER}->getStatements(@_);}
sub setProofViews {return shift->{QRESULTS}->setProofViews(@_);}
sub refreshQueryResults {return shift->{QRESULTS}->refresh(@_);}
sub showQueryResultsWindow {return shift->{QRESULTS}->showWindow(@_);}
sub refreshList {return shift->{STATEMENTS}->refreshList(@_);}
sub findTriple {return shift->{STATEMENTS}->findTriple(@_);}
sub clearRenderedTriples {return shift->{GRAPH}->clearRenderedTriples(@_);}
sub addRenderedTriple {return shift->{GRAPH}->addRenderedTriple(@_);}
sub graphNode {return shift->{GRAPH}->graphNode(@_);}
sub graphStatement {return shift->{GRAPH}->graphStatement(@_);}
sub showGraphWindow {return shift->{GRAPH}->showWindow(@_);}

sub show {
    my ($self, $prefix, $namespaceHandler) = @_; $prefix = '  ' if (!defined $prefix);
    my $ret = '';
    $ret .= $prefix.$self."\n";
    $ret .= $prefix.'RDF_PREFIX: '.$self->{RDF_PREFIX}."\n";
    $ret .= $prefix.'CURRENT_SOURCE_ATTRIBUTION: '.$self->{CURRENT_SOURCE_ATTRIBUTION}->dump."\n" if ($self->{CURRENT_SOURCE_ATTRIBUTION});
    $ret .= $prefix."NAMESPACES:\n";map {$ret .= $prefix.'  '.$_.': '.$self->{NAMESPACES}{$_}."\n"} keys %{$self->{NAMESPACES}};
    $ret .= $prefix."CONTAINERS:\n";map {$ret .= $self->{CONTAINERS}{$_}->show($prefix.'  ')} keys %{$self->{CONTAINERS}};
    $ret .= $prefix."ABOUT_EACH:\n";map {$ret .= $prefix.'  '.$_.': ('.join(',', @{$self->{ABOUT_EACH}{$_}}).")\n"} keys %{$self->{ABOUT_EACH}};
    $ret .= $prefix."ABOUT_EACH_PREFIX:\n";map {$ret .= $prefix.'  '.$_.': ('.join(',', @{$self->{ABOUT_EACH_PREFIX}{$_}}).")\n"} keys %{$self->{ABOUT_EACH_PREFIX}};
    $ret .= $prefix."TRIPLES:\n";
    $ret .= $self->showTriples($prefix.'  ', $namespaceHandler);
    return $ret;
}

package W3C::Rdf::RdfVisualizer::ClientWindow;
# pure virtuals
sub populateWindow {die}

sub new {
    my ($proto, $parm) = @_;
    my $class = ref($proto) || $proto;
    my $self;
    $self = $parm;
    bless ($self, $class);
    return $self;
}

package W3C::Rdf::RdfVisualizer::StatementList;

sub makeWindow {
    my ($self, $listFrame, $searchFrame) = @_;

    # search prompt
    $searchFrame->Label(-text => 'Look for: ')->pack(-side => 'left');
    $self->{SEARCH} = $searchFrame->Scrolled('Entry', -scrollbars => 'os')
	->pack(-side => 'left', -anchor => 'w', -fill => 'x', -expand => 1);
    $self->{SEARCH}->bind("<Return>", sub {$self->search(0)});
    $self->{SEARCH}->bind("<Up>", sub {$self->search(1)});
    $self->{SEARCH}->bind("<Down>", sub {$self->search(0)});

    # statement list
    $self->{TRIPLE_LIST} = &{$self->{-matrixWidgetFactory}} (({-parentWindow => $listFrame, 
							       -command => sub {$self->triplesPick(@_);}, 
							       -columnLabels => ['#','attrib','','predicate',
										 'subject','object','','reifId','as']}));

    $self->{TRIPLE_LIST}->pack('-expand' => 1, '-fill' => 'both');
    $self->{TRIPLE_LIST}->bind("<Control-Alt-Up>", sub {$self->search(1)});
    $self->{TRIPLE_LIST}->bind("<Control-Alt-Down>", sub {$self->search(0)});
}

sub refreshList {
    my ($self) = @_;
    $self->{TRIPLE_LIST}->startRows;
    my $i = 0;
    $self->{-windowSet}->clearRenderedTriples;
    $self->{BY_INDEX} = undef;
    my @statements = $self->sortStatements($self->{-windowSet}->getStatements);
    for (my $i = 0; $i < @statements; $i++) {
	my $statement = $statements[$i];
	my ($enabled, $segments) = $self->{-windowSet}->renderStatementSegments($statement);
	next if (!$enabled);
	$self->{TRIPLE_LIST}->addRow([$i, @$segments, '', '']);
	$self->{-windowSet}->addRenderedTriple('triple_'.$statement, $i, $statement);
	$self->{BY_INDEX}[$i] = [join (' ', @$segments), $statement];
    }
    my $windowWidth = $self->{TRIPLE_LIST}->endRows;
#    $self->{WINDOW}->configure('-width' => $windowWidth); # @@@ - no worky
}

sub search {
    my ($self, $up) = @_;
    my $search = $self->{SEARCH};
    my $searchString = $search->get;
    return if ($searchString eq '');
    my $lb = $self->{TRIPLE_LIST};
    if (!$up) {
	my $start = $self->{SEARCH_INDEX} == $lb->size - 1 || !defined $self->{SEARCH_INDEX} ? 0 : $self->{SEARCH_INDEX}+1;
	my $end = $lb->size - 1;
	my $i;
      SEARCH_DOWN:
	for ($i = $start; $i <= $end; $i++) {
	    my $entry = $lb->get($i);
	    if ($entry =~ m/$searchString/) {
		$lb->highlight($i);
		$self->{SEARCH_INDEX} = $i;
		$self->triplesPick($i, 0);
		return;
	    }
	}
	if ($i == $lb->size) {
	    $start = 0;
	    $end = $self->{SEARCH_INDEX};
	    goto SEARCH_DOWN;
	}
    } else {
	my $start = $self->{SEARCH_INDEX} == 0 || !defined $self->{SEARCH_INDEX} ? $lb->size - 1 : $self->{SEARCH_INDEX}-1;
	my $end = 0;
	my $i;
      SEARCH_UP:
	for ($i = $start; $i >= $end; $i--) {
	    my $entry = $lb->get($i);
	    if ($entry =~ m/$searchString/) {
		$lb->highlight($i);
		$self->{SEARCH_INDEX} = $i;
		$self->triplesPick($i,0);
		return;
	    }
	}
	if ($i == -1) {
	    $start = $lb->size - 1;
	    $end = $self->{SEARCH_INDEX}+1;
	    goto SEARCH_UP;
	}
    }
    $lb->bell;
#    $search->flash; # find a button to do the flashing
#    print "not found\n";
}

sub sortStatements {
    my ($self) = (shift);
    return @_;
    eval { # perl 5.004_04 seems to have a prob with this:
	return sort {$self->{-windowSet}->renderNode($a->getPredicate, $a->getAttribution) cmp 
			 $self->{-windowSet}->renderNode($b->getPredicate, $b->getAttribution);} @_;
    }; # so I provide this alternative:
    return @_;
}

sub triplesPick {
    my ($self, $row, $col) = @_;
    $self->{SEARCH_INDEX} = $row;
    my $statement = $self->{BY_INDEX}[$row]->[1];
    if ($col == 4) {
	$self->{-windowSet}->graphNode($statement->getSubject);
    } elsif ($col == 5) {
	$self->{-windowSet}->graphNode($statement->getObject);
    } else {
	$self->{-windowSet}->graphStatement($statement, 1);
    }
    $self->{-windowSet}->showGraphWindow;
}

sub findTriple { # used by Graph::graphPick
    my ($self, $index) = @_;
    $self->{TRIPLE_LIST}->highlight($index);
}

package W3C::Rdf::RdfVisualizer::MainWindow;
# pure virtuals
sub getTitle {die}

sub makeWindow {
    my ($self, $parentWindow) = @_;
    $self->{PARENT_WINDOW} = $parentWindow; # @@@
    $self->{WINDOW} = $parentWindow->Toplevel;
    $self->{WINDOW}->title($self->getTitle);
    $self->{WINDOW}->protocol('WM_DELETE_WINDOW', sub {$self->destroyWindow});
    $self->{WINDOW}->withdraw if (!$self->{SHOW});
    $self->populateWindow;
}
sub hideWindow {$_[0]->{SHOW} = 0; $_[0]->viewWindow;}
sub showWindow {$_[0]->{SHOW} = 1; $_[0]->viewWindow;}
sub viewWindow {
    my ($self) = @_;
    if ($self->{SHOW}) {
	$self->{WINDOW}->deiconify();
	$self->{WINDOW}->raise();
    } else {
	$self->{WINDOW}->withdraw;
    }
}
sub destroyWindow {
    my ($self) = @_;
    $self->{WINDOW}->destroy;
    $self->{SHOW} = 0;
    $self->makeWindow($self->{PARENT_WINDOW});
}
sub getShowVariable {return \ $_[0]->{SHOW};}

package W3C::Rdf::RdfVisualizer::Graph;
sub new {
    my ($proto, $parms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new($parms);
    bless ($self, $class);
    $self->{GRAPH_MASTER} = new W3C::Rdf::RdfVisualizer::GraphMaster({%$parms});
    return $self;
}

# overloaded virtuals
sub getTitle {return 'Graph';}
sub populateWindow {
    my ($self) = @_;
    my $parentWindow = $self->{WINDOW}->Frame(-relief => 'groove', -bd => 2)->pack(-fill => 'x');
    # control buttons
    $self->{GRAPH_MASTER}->makeWindow($parentWindow);
    $parentWindow->Button(-text => 'Hide', -command => sub {$self->hideWindow})->pack(-side => 'right');
    $self->{CANVAS} = $self->{WINDOW}->Scrolled('Canvas', -width=>$X, -height=>($Y * 5)/4) ->pack(-fill => 'both', -expand => 1);
    $self->{CANVAS}->Subwidget('canvas')->Tk::bind("<Button-1>", sub {$self->graphPick});
}

# support functions
sub getCanvas {$_[0]->{CANVAS};}
sub clearCanvas {$_[0]->{CANVAS}->delete('stuff');}
sub getLastCurrent {$_[0]->{LAST_CURRENT};}
sub graphPick {
    my ($self, $overrideCurrent) = @_;
    my $graph = $self->getCanvas;
    my @current = $graph->gettags('current');
    @current = @$overrideCurrent if ($overrideCurrent);
    $self->{LAST_CURRENT} = \@current;
    return if ($current[0] !~ m/\A(triple|node|nav)_(.*)\Z/);
    if ($1 eq 'triple') {
	# statements->findTriple(index of the selected statement (may be undef))
	$self->{-windowSet}->findTriple($self->{BY_TRIPLE}{$current[0]}->[0]);
    } elsif ($1 eq 'node') {
	$self->graphNode($self->{STR_TO_OBJ}{$current[0]});
    } elsif ($1 eq 'nav') {
	$self->{GRAPH_MASTER}->onArrow($2 eq 'back' ? -1 : 1)
    }
}

sub clearRenderedTriples {
    my ($self) = @_;
    $self->{BY_TRIPLE} = undef;
}

sub addRenderedTriple {
    my ($self, $tag, $index, $statement) = @_;
    $self->{BY_TRIPLE}{$tag} = [$index, $statement];
}

sub graphStatement {
    my ($self, $statement) = @_;
    my ($enabled, $segments) = $self->{-windowSet}->renderStatementSegments($statement);
    my $rendered = '('.$segments->[2].' '.$segments->[3].' '.$segments->[4].')';
    my $view = new W3C::Rdf::RdfVisualizer::GraphStatementView({-windowSet => $self->{-windowSet}, 
								-string => $rendered, 
								-statement => $statement, 
								-graph => $self});
    $self->{GRAPH_MASTER}->add($view);
    $view->renderGraphView;
}

sub graphNode {
    my ($self, $node, $attribution) = @_;
    my $rendered = $self->{-windowSet}->renderNode($node, $attribution);
    my $view = $self->{GRAPH_MASTER}->getByString($rendered);
    if (!$view) {
	$view = new W3C::Rdf::RdfVisualizer::GraphNodeView({-windowSet => $self->{-windowSet}, 
							    -string => $rendered, 
							    -node => $node, 
							    -graph => $self});
	$self->{GRAPH_MASTER}->add($view);
    }
    $view->renderGraphView;
}

sub graphTriples {
    my ($self, $triples, $fullScreen, $top) = @_;
    my $graph = $self->getCanvas;
    my $tripleCount = $#$triples - $[ + 1;
    my $step = $X / ($tripleCount + 1);
    my $x = $step;
    my ($yTop, $yMid, $yBot) = $fullScreen ? ($TS+1, $Y/2, $Y-$TS) : $top ? ($TS+1, $Y/4, $Y/2) : ($Y/2, (3*$Y)/4, $Y-$TS);
    for (my $i = 0; $i < $tripleCount; $i++) {
	my $triple = $triples->[$i];
	my ($predicate, $subject, $object, $attrib) = 
	    ($triple->getPredicate, $triple->getSubject, $triple->getObject, $triple->getAttribution);
	my ($predicateShort, $subjectShort, $objectShort, $sourceAttrib) = 
	    ($self->{-windowSet}->renderNode($predicate, $attrib), 
	     $self->{-windowSet}->renderNode($subject, $attrib), 
	     $self->{-windowSet}->renderNode($object, $attrib), 
	     $attrib);
	my $node = $top ? $subject : $object;
	my $entryNode = $self->{GRAPH_MASTER}->getPrevious->matchesNode($node);
	my $exitNode  = $self->{GRAPH_MASTER}->getNext->matchesNode($node);
	my $lineColor = $entryNode ? 'red' : $exitNode ? 'green' : 'black';
	my $lineTag = $entryNode ? 'nav_back' : $exitNode ? 'nav_forward' : 'inert';
#	$graph->createOval(60,1, 140,31);
	my $xSubjCenter = ($fullScreen||!$top)?$X/2:$x;
	my $xPredCenter = ($fullScreen)?$X/2:$x;
	my $xObjCenter = ($fullScreen||$top)?$X/2:$x;
	my ($subjectTag, $objectTag, $tripleTag) = ('node_'.$subject, 'node_'.$object, 'triple_'.$triple);
	$self->{STR_TO_OBJ}{$subjectTag} = $subject;
	$self->{STR_TO_OBJ}{$objectTag} = $object;
	$self->{STR_TO_OBJ}{$tripleTag} = $triple;
	# subject
	$graph->createText($xSubjCenter,$yTop, 
			   -justify => 'center', 
			   -text => $subjectShort, 
			   -fill => 'blue', 
			   -tags => [$subjectTag, 'stuff']);
	$graph->createLine($xSubjCenter,$yTop+$TS, $xPredCenter,$yMid-$TS, 
			   -tags => [$lineTag, 'stuff'], 
			   -fill => $lineColor);
	# predicate
	$graph->createText($xPredCenter,$yMid, 
			   -justify => 'left', 
			   -text => $predicateShort, 
			   -tags => [$tripleTag, 'stuff']);
	$graph->createLine($xPredCenter,$yMid+$TS, $xObjCenter,$yBot-$TS, 
			   -tags => [$lineTag, 'stuff'], 
			   -fill => $lineColor);
	# arrow
	$graph->createLine($xObjCenter-$AW,$yBot-$TS-$AH, $xObjCenter,$yBot-$TS, 
			   -tags => [$lineTag, 'stuff'], 
			   -fill => $lineColor);
	$graph->createLine($xObjCenter+$AW,$yBot-$TS-$AH-1, $xObjCenter,$yBot-$TS, 
			   -tags => [$lineTag, 'stuff'], 
			   -fill => $lineColor);
	# object
	$graph->createText($xObjCenter,$yBot, 
			   -justify => 'center', 
			   -text => $objectShort, 
			   -fill => 'blue', 
			   -tags => [$objectTag, 'stuff']);
#	$graph->createOval(60,170, 140,199);
	$x += $step;
	if (!$fullScreen) {
	    if ($top) {
		$yTop+=$TS;
	    } else {
		$yBot+=$TS;
	    }
	    $yMid+=$TS;
	}
    }
}

package W3C::Rdf::RdfVisualizer::QueryResults;
# overloaded virtuals
sub getTitle {return 'Query Results';}
sub populateWindow {
    my ($self) = @_;
    $self->{WINDOW}->Button(-text => 'Hide', -command => sub {$self->hideWindow})->pack;
}

sub destroyWindow {
    my ($self) = @_;
    $self->SUPER::destroyWindow;
    $self->refresh;
}

# QueryResults-specific methods
sub setProofViews {
    my ($self, $columnlabels, $proofViews) = @_;
    $self->{COLUMN_LABELS} = $columnlabels;
    $self->{PROOF_VIEWS} = $proofViews;
}

sub refresh {
    my ($self) = @_;

#    $self->{COLUMNS}->destroy if ($self->{COLUMNS});
    $self->{COLUMNS} = &{$self->{-matrixWidgetFactory}} (({-parentWindow => $self->{WINDOW}, 
							   -command => sub {$self->pick(@_);}, 
							   -columnLabels => ['#', @{$self->{COLUMN_LABELS}}]}));
    $self->{COLUMNS}->startRows;
    for (my $i = 0; $i < @{$self->{PROOF_VIEWS}}; $i++) {
	$self->{COLUMNS}->addRow([$i, $self->{PROOF_VIEWS}[$i]->renderData]);
    }

    my $windowWidth = $self->{COLUMNS}->endRows;
    $self->{WINDOW}->configure('-width' => $windowWidth); # @@@ - no worky
    $self->{COLUMNS}->pack('-expand' => 1, '-fill' => 'both');
}

sub pick {
    my ($self, $row, $column) = @_;
    my $view = $self->{PROOF_VIEWS}[$row];
    $self->{-windowSet}->addView($view);
    if ($column > 0) {
	my $datum = $view->{-data}[$column-1];
	$self->{-windowSet}->graphNode($datum);
	$self->{-windowSet}->showGraphWindow;
    }
    $self->{-windowSet}->refreshList;
}

package W3C::Rdf::RdfVisualizer::ListMaster;

sub new {
    my ($proto, $parms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new($parms);
    bless ($self, $class);
    $self->{INDEX} = -1;
    $self->{STACK} = [];
    $self->{BY_STRING} = {};
    return $self;
}

# virtual functions
sub makeWindow { # ClientWindow
    my ($self, $parentWindow) = @_;
    $self->{ENTRY} = $self->makeEntry($parentWindow);
    $self->{ENTRY}->bind("<Return>", sub {$self->onReturn});
    $self->{ENTRY}->bind("<Up>", sub {$self->onArrow(-1)});
    $self->{ENTRY}->bind("<Down>", sub {$self->onArrow(1)});
#    $self->{LEFT} = $parentWindow->Button(-text => "<-", -underline => 0, -state => 'disabled', 
#				-command => sub {$self->nav(1)}, -background => 'grey')->pack(-side => 'left');
#    $self->{RIGHT} = $parentWindow->Button(-text => "->",  -underline => 1, -state => 'disabled', 
#				 -command => sub {$self->nav(1)}, -background => 'grey')->pack(-side => 'left');
}

# control window event redirectees
sub add {
    my ($self, $view) = @_;
#    return if (grep {$_->getString eq $view->getString} @{$self->{VIEW}});
    $self->{TEXT_VARIABLE} = $view->getString;
    push (@{$self->{STACK}}, $view);
    $self->{INDEX} = scalar @{$self->{STACK}} - 1;
    $self->{BY_STRING}{$view->getString} = $view;
}

sub getByString {
    my ($self, $string) = @_;
    return $self->{BY_STRING}{$string};
}

sub onArrow {
    my ($self, $direc) = @_;
    return if (@{$self->{STACK}} < 1);
    $self->{INDEX} += $direc;
    $self->{INDEX} += @{$self->{STACK}} if ($self->{INDEX} < 0);
    $self->{INDEX} -= @{$self->{STACK}} if ($self->{INDEX} >= @{$self->{STACK}});
    $self->{TEXT_VARIABLE} = $self->{STACK}[$self->{INDEX}]->getString;
    $self->{STACK}[$self->{INDEX}]->activate($self);
    $self->{-windowSet}->refreshList;
}

sub getPrevious {
    my ($self) = @_;
    return $self->{INDEX} > 0                  ? $self->{STACK}[$self->{INDEX} - 1] : $self->{STACK}[$#{$self->{STACK}}];
}
sub getNext {
    my ($self) = @_;
    return $self->{INDEX} < $#{$self->{STACK}} ? $self->{STACK}[$self->{INDEX} + 1] : $self->{STACK}[0];
}

sub imActive {
    my ($self, $newActive) = @_;
    $self->{INDEX} = (grep {return $_ == $newActive;} @{$self->{STACK}})[0];
    die if (!defined $self->{INDEX});
    $self->{TEXT_VARIABLE} = $newActive->getString;
}

package W3C::Rdf::RdfVisualizer::ViewMaster;

# virtual functions
sub makeEntry { # ListMaster downcall
    my ($self, $parentWindow) = @_;
    $parentWindow->Label(-text => 'Now playing: ')->pack(-side => 'left');
    return $parentWindow->Scrolled('Entry', -scrollbars => 'os', -textvariable => \ $self->{TEXT_VARIABLE}, -width => 25)
	->pack(-side => 'left', -anchor => 'w');
}

sub passed {
    my ($self, $viewName) = @_;
    $self->{TEXT_VARIABLE} = $viewName;
    $self->onReturn;
}

sub onReturn {
    my ($self) = @_;
    my @sources = split (/\s/, $self->{TEXT_VARIABLE});
    my (@nextAttribs, @nextSources);
    foreach my $source (@sources) {
	if (exists $self->{-windowSet}->{-attributionsByInputSource}{$source}) {
	    push (@nextAttribs, $self->{-windowSet}->{-attributionsByInputSource}{$source});
	    push (@nextSources, $source);
	} elsif ($source !~ /^\s*$/) {
	    $self->{ENTRY}->bell;
	}
    }
    if (@nextAttribs) {
	$self->{TEXT_VARIABLE} = join (' ', @nextSources);
	if (grep {$_->getString eq $self->{TEXT_VARIABLE}} @{$self->{STACK}}) {
	} else {
	    my $currentView = new W3C::Rdf::RdfVisualizer::SourceView({VIEW_MASTER => $self->{VIEW_MASTER}, 
								       -windowSet => $self->{-windowSet}, 
								       -string => $self->{TEXT_VARIABLE}, -statements => undef, 
								       -attribs => \@nextAttribs});
	    $self->add($currentView);
	}
	$self->{-windowSet}->refreshList;
    }
}

sub getStatements {
    my ($self) = @_;
    return $self->{INDEX} > -1 ? $self->{STACK}[$self->{INDEX}]->getStatements : ();
}

package W3C::Rdf::RdfVisualizer::View;
# pure virtuals
sub refresh {die}
sub getStatements {die}

sub new {
    my ($proto, $parms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $parms;
    bless ($self, $class);
    return $self;
}

sub getString {
    my ($self) = @_;
    return $self->{-string};
}

sub activate {
    my ($self, $caller) = @_;
}

package W3C::Rdf::RdfVisualizer::SourceView;
sub getStatements {
    my ($self) = @_;
#    may want a parent view - need to work out interface to view selection
    return $self->{-windowSet}->{-RdfDB}->triplesMatching(undef, [[undef,undef,undef]], 
							  {$self->{-windowSet}->getQueryOptions, 
							   -attributions=>$self->{-attribs}});
}

package W3C::Rdf::RdfVisualizer::QueryView;
sub activate {
    my ($self, $caller) = @_;
    if ($caller == $self->{-queryMaster}) {
	$self->{-viewMaster}->imActive($self);
    } elsif ($caller == $self->{-viewMaster}) {
	$self->{-queryMaster}->imActive($self);
    } else {
	die "$caller has no damn right to activate $self";
    }
}
sub getStatements {
    my ($self) = @_;
    my ($nodes, $selects, $messages, $proofs) = 
	$self->{-windowSet}->{-RdfDB}->guessQuery($self->{-string}, $self->{-systemId});
    return $self->processAlgernonResults($nodes, $selects, $messages, $proofs);
}

sub processAlgernonResults {
    my ($self, $nodes, $selects, $messages, $proofs) = @_;
    my (%ret, @proofViews);
    for (my $i = 0; $i < @$nodes; $i++) {
	my $row = $nodes->[$i];
	my $rowProofs = $proofs->[$i];
	foreach my $rowProof (@$rowProofs) {
	    $ret{$rowProof} = $rowProof;
	}
	my $proofView = new W3C::Rdf::RdfVisualizer::ProofView({VIEW_MASTER => $self->{VIEW_MASTER}, 
								-windowSet => $self->{-windowSet}, 
								-string => $self->{TEXT_VARIABLE}.'.'.$i, 
								-statements => $rowProofs, -data => $row});
	push (@proofViews, $proofView);
    }
    $self->{-windowSet}->setProofViews($selects, \@proofViews);
    return values %ret;
}

package W3C::Rdf::RdfVisualizer::ProofView;
sub getStatements { # proof statements are static (not re-queried) for simplicity
    my ($self) = @_;
    return @{$self->{-statements}};
}
sub renderData {
    my ($self) = @_;
    my (@data);
    for (my $i = 0; $i < @{$self->{-data}}; $i++) {
	my $rendered = $self->{-windowSet}->renderNode($self->{-data}[$i], $self->{-statements}[0]->getAttribution, 
						       {-stringTrunc => 30, -truncFlag => '...'});
	push (@data, $rendered);
    }
    return @data;
}

package W3C::Rdf::RdfVisualizer::GraphView;
sub renderData {
    my ($self) = @_;
    my (@data);
    for (my $i = 0; $i < @{$self->{-data}}; $i++) {
	my $rendered = $self->{-windowSet}->renderNode($self->{-data}[$i], $self->{-statements}[0]->getAttribution, 
						       {-stringTrunc => 30, -truncFlag => '...'});
	push (@data, $rendered);
    }
    return @data;
}

sub activate {
    my ($self) = @_;
    $self->renderGraphView;
}

sub matchesNode {die}

package W3C::Rdf::RdfVisualizer::GraphStatementView;
sub renderGraphView {
    my ($self) = @_;
    $self->{-graph}->clearCanvas;
    $self->{-graph}->graphTriples([$self->{-statement}], 1);
}

sub matchesNode {
    my ($self, $node) = @_;
    return 1 if ($self->{-statement}->getSubject == $node);
    return 1 if ($self->{-statement}->getObject == $node);
    return 0;
}

package W3C::Rdf::RdfVisualizer::GraphNodeView;
sub renderGraphView {
    my ($self) = @_;
    $self->{-graph}->clearCanvas;

    my @triples;
    @triples = $self->{-windowSet}->{-RdfDB}->triplesMatching(undef, [[undef, undef, $self->{-node}]], 
							      {$self->{-windowSet}->getQueryOptions});
    $self->{-graph}->graphTriples(\@triples, 0, 1);
    @triples = $self->{-windowSet}->{-RdfDB}->triplesMatching(undef, [[undef, $self->{-node}, undef]], 
							      {$self->{-windowSet}->getQueryOptions});
    $self->{-graph}->graphTriples(\@triples, 0, 0);
}

sub matchesNode {
    my ($self, $node) = @_;
    return $self->{-node} == $node;
}

package W3C::Rdf::RdfVisualizer::QueryMaster;

# virtual functions
sub makeEntry { # ListMaster downcall
    my ($self, $parentWindow) = @_;
    $parentWindow->Label(-text => 'Query: ')->pack(-side => 'left');
    return $parentWindow->Scrolled('Entry', -scrollbars => 'os', -textvariable => \ $self->{TEXT_VARIABLE})->pack(-fill => 'x');
}

sub passed {
    my ($self, $query) = @_;
    $self->{TEXT_VARIABLE} = $query;
    $self->onReturn;
}

sub onReturn {
    my ($self) = @_;
    eval {
	if (grep {$_->getString eq $self->{TEXT_VARIABLE}} @{$self->{STACK}}) {
	} else {
	    my ($nodes, $selects, $messages, $proofs) = 
		$self->{-windowSet}->{-RdfDB}->guessQuery($self->{TEXT_VARIABLE}, $self->{-systemId});
	    my $viewName = 'query-'.($self->{INDEX}+1);
	    my $currentView = new W3C::Rdf::RdfVisualizer::QueryView({-viewMaster => $self->{-windowSet}->{VIEW_MASTER}, 
								      -windowSet => $self->{-windowSet}, 
								      -string => $self->{TEXT_VARIABLE}, 
								      -statements => undef, 
								      -query => $self->{TEXT_VARIABLE}, 
								      -viewTitle => $viewName, 
								      -queryMaster => $self});
	    $self->add($currentView);
	    $self->{-windowSet}->addView($currentView);
	}
	$self->{-windowSet}->refreshList;
	$self->{-windowSet}->refreshQueryResults;
	$self->{-windowSet}->showQueryResultsWindow;
    }; if ($@) {
	$self->{ENTRY}->bell;
    }
}

sub onArrow {
    my ($self, $direc) = @_;
    return if (@{$self->{STACK}} < 1);
    $self->SUPER::onArrow($direc);
    $self->{-windowSet}->refreshQueryResults;
}

package W3C::Rdf::RdfVisualizer::GraphMaster;

# virtual functions
sub makeEntry { # ListMaster downcall
    my ($self, $parentWindow) = @_;
    $parentWindow->Label(-text => 'Node: ')->pack(-side => 'left');
    return $parentWindow->Scrolled('Entry', -scrollbars => 'os', -textvariable => \ $self->{TEXT_VARIABLE})
	->pack(-side => 'left', -fill => 'x', -expand => 1, -padx => 10);
}

sub passed {
    my ($self, $node) = @_;
    $self->{TEXT_VARIABLE} = $node;
    $self->onReturn;
}

sub onReturn {
    my ($self) = @_;
    $self->{ENTRY}->bell;
}

package W3C::Rdf::RdfVisualizer;

1;


__END__

=head1 NAME

W3C::Rdf::RdfVisualizer - PerlTk tool to view an RdfDB

=head1 SYNOPSIS

  use W3C::Rdf::RdfDB;
  use W3C::Rdf::RdfVisualizer;
  my $rdfDB = new W3C::Rdf::RdfDB;
  my $rdfVisualizer = new W3C::Rdf::RdfVisualizer($rdfDB);
  my $desc = $rdfDB->addDescription(...);
  $desc->addPredicateObject($predicate, $object);
  $rdfDB->browse;

=head1 DESCRIPTION

RdfVisualizer was a quick hack to provide a Tk interface to browse the
triples in an RdfDB. It provides a list of triples and a canvas where
the user may click on a triple and explore the surrounding nodes. It's
also my first PerlTk app so if you notice something that may be awry,
assume it is and let me know how to improve it.

This module is part of the W3C::Rdf CPAN module.

=head2 Hack

It's pricipal shortcoming is that I didn't use a defined interface to
NamespaceHandler but instead used its internal variables.

=head1 AUTHOR

Eric Prud\'hommeaux <eric@w3.org>

=head1 SEE ALSO

W3C::Rdf::RdfParser(3) perl(1).

=cut
